/* environment.c -- An "Environment" is a set of variables, that is, a mapping from symbols to values */

#include <stdlib.h>
#include <stdio.h>
#include "locksim.h"

struct Environment* create_environment(void) {
    struct Environment* p = malloc(sizeof(struct Environment));
    if (p == NULL)
        fatal_error("Memory full.");
    p->variables = NULL;
    p->nr_variables = 0;
    p->variable_table_length = 0;
    return p;
}

void destroy_environment(struct Environment* e) {
    free(e->variables);
    free(e);
}

/* Returns the variable table index for a symbol, or -1 if no such variable */
/* The variable table index is not (necessarily) the same as the symbol table index */
static int find_variable_place(struct Environment* environment, int symbol_index) {
    for (int i = environment->nr_variables - 1; i >= 0; --i) {
        if (environment->variables[i].symbol_index == symbol_index)
            return i;
    }
    return -1;
}

/* Inserts a variable into the variable table, and returns its index */
static int insert_variable(struct Environment* environment, int symbol_index) {
    if (find_variable_place(environment, symbol_index) != -1)
        fatal_error("Tried to insert a variable that already exists.\n");
    if (environment->nr_variables == environment->variable_table_length) {
        if (environment->variable_table_length == 0)
            environment->variable_table_length = 8;
        else
            environment->variable_table_length *= 4;
        environment->variables =
            realloc(environment->variables,
                environment->variable_table_length * sizeof(struct Variable));
        if (environment->variables == NULL)
            fatal_error("Memory full.");
    }
    environment->variables[environment->nr_variables].symbol_index = symbol_index;
    environment->variables[environment->nr_variables].value = -1000;
    return environment->nr_variables++;
}

int get_variable_value(struct Environment* environment, int symbol_index) {
    int place = find_variable_place(environment, symbol_index);
    if (place == -1)
        fatal_error("No such variable: %s", get_symbol_name_unsafe(symbol_index));
    return environment->variables[place].value;
}

void set_variable_value(struct Environment* environment, int symbol_index, int value) {
    int place = find_variable_place(environment, symbol_index);
    if (place == -1)
        place = insert_variable(environment, symbol_index);
    environment->variables[place].value = value;
}

void show_environment(struct Environment* environment) {
    printf("Environment with %d variable%s:\n", environment->nr_variables, environment->nr_variables == 1 ? "" : "s");
    for (int i = 0; i < environment->nr_variables; ++i) {
        printf("    %d. %s = %d\n",
            i,
            get_symbol_name_unsafe(environment->variables[i].symbol_index),
            environment->variables[i].value);
    }
}

void show_environment_compact(struct Environment* environment) {
    if (environment->nr_variables == 0)
        printf("No variables");
    else
        printf("%d variable%s: ", environment->nr_variables, environment->nr_variables == 1 ? "" : "s");
    for (int i = 0; i < environment->nr_variables; ++i) {
        printf("%s = %d",
            get_symbol_name_unsafe(environment->variables[i].symbol_index),
            environment->variables[i].value);
        if (i < environment->nr_variables - 1)
            printf(", ");
    }
    printf("\n");
}

int equal_environments(struct Environment* e1, struct Environment* e2) {
    if (e1->nr_variables != e2->nr_variables)
        return 0;
    else {
        for (int i = 0; i < e1->nr_variables; ++i) {
            int other_value = get_variable_value(e2, e1->variables[i].symbol_index);
            if (e1->variables[i].value != other_value)
                return 0;
        }
        return 1;
    }
}
