/* symbols.c -- The symbol table keeps track of identifiers, such as variable names */

#ifdef _WIN32
#define _CRT_SECURE_NO_WARNINGS
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include "locksim.h"

static struct Symbol* symbol_table = NULL;
static int nr_symbols = 0;
static int symbol_table_length = 0;

static int my_strcasecmp(char* s1, char* s2) {
    while (*s1) {
        if (!*s2)
            return 1;
        char c1 = (unsigned char)*s1;
        char c2 = (unsigned char)*s2;
        if (isupper(c1))
            c1 = tolower(c1);
        if (isupper(c2))
            c2 = tolower(c2);
        if (c1 == c2)
            ;
        else if (c1 < c2)
            return -1;
        else
            return 1;
        ++s1;
        ++s2;
    }
    if (*s2)
        return -1;
    else
        return 0;
}

/* Returns symbol table index for a string, or -1 if no such symbol */
static int lookup_symbol(char* name) {
    for (int i = nr_symbols - 1; i >= 0; --i) {
        // if (strcmp(symbol_table[i].name, name) == 0)
        if (my_strcasecmp(symbol_table[i].name, name) == 0)
            return i;
    }
    return -1;
}

/* Inserts a symbol into the symbol table, and returns its index */
static int insert_symbol(char* name, int token_type) {
    if (lookup_symbol(name) != -1)
        fatal_error("Tried to insert a symbol that already exists.\n");
    int name_length = strlen(name);
    if (nr_symbols == symbol_table_length) {
        if (symbol_table_length == 0)
            symbol_table_length = 16;
        else
            symbol_table_length *= 4;
        symbol_table = realloc(symbol_table, symbol_table_length * sizeof(struct Symbol));
        if (symbol_table == NULL)
            fatal_error("Memory full.");
    }
    char* copy = malloc(name_length + 1);
    if (copy == NULL)
        fatal_error("Memory full.");
    strcpy(copy, name);
    symbol_table[nr_symbols].name = copy;
    symbol_table[nr_symbols].token_type = token_type;
    return nr_symbols++;
}

int lookup_or_insert_symbol(char* name, int token_type) {
    int result = lookup_symbol(name);
    if (result == -1)
        result = insert_symbol(name, token_type);
    return result;
}

/* "Unsafe" = the returned string isn't copied, but a pointer into the symbol table's data */
char* get_symbol_name_unsafe(int symbol_index) {
    if (symbol_index < 0 || symbol_index >= nr_symbols)
        fatal_error("Invalid symbol index.");
    return symbol_table[symbol_index].name;
}

int get_symbol_type(int symbol_index) {
    if (symbol_index < 0 || symbol_index >= nr_symbols)
        fatal_error("Invalid symbol index.");
    return symbol_table[symbol_index].token_type;
}

void init_symbols(void) {
    insert_symbol("transaction", TRANSACTION);
    insert_symbol("trans", TRANSACTION);
    insert_symbol("read", READ);
    insert_symbol("write", WRITE);
    insert_symbol("done", DONE);
    insert_symbol("go", DONE);
    insert_symbol("lock", LOCK);
    insert_symbol("readlock", READLOCK);
    insert_symbol("writelock", WRITELOCK);
    insert_symbol("unlock", UNLOCK);
}
