/* trees.c -- Syntax trees */

#include <stdlib.h>
#include <stdio.h>
#include <assert.h>
#include "locksim.h"

struct TreeNode* create_node(int type, struct TreeNode* arg1, struct TreeNode* arg2) {
    struct TreeNode* p = malloc(sizeof(struct TreeNode));
    if (p == NULL)
        fatal_error("Memory full.");
    p->type = type;
    p->args[0] = arg1;
    p->args[1] = arg2;
    return p;
}

struct TreeNode* create_number_node(int value) {
    struct TreeNode* p = malloc(sizeof(struct TreeNode));
    if (p == NULL)
        fatal_error("Memory full.");
    p->type = NUMBER;
    p->leaf_value = value;
    return p;
}

struct TreeNode* create_id_node(int symbol_index) {
    struct TreeNode* p = malloc(sizeof(struct TreeNode));
    if (p == NULL)
        fatal_error("Memory full.");
    p->type = ID;
    p->leaf_value = symbol_index;
    return p;
}

struct TreeNode* car(struct TreeNode* cons_node) {
    assert(cons_node->type == ';');
    return cons_node->args[0];
}

struct TreeNode* cdr(struct TreeNode* cons_node) {
    assert(cons_node->type == ';');
    return cons_node->args[1];
}

void rplacd(struct TreeNode* cons_node, struct TreeNode* new_cdr) {
    assert(cons_node->type == ';');
    cons_node->args[1] = new_cdr;
}

int list_length(struct TreeNode* cons_node) {
    int length = 0;
    while (cons_node != NULL) {
        assert(cons_node->type == ';');
        ++length;
        cons_node = cdr(cons_node);
    }
    return length;
}

static void print_tree(struct TreeNode* p, int level) {
    if (p == NULL)
        ;
    else if (p->type == ID) {
        printf("%*s", 2 * level, "");
        printf("%s\n", get_symbol_name_unsafe(p->leaf_value));
    }
    else if (p->type == NUMBER) {
        printf("%*s", 2 * level, "");
        printf("%d\n", p->leaf_value);
    }
    else if (p->type == ';') {
        printf("%*s", 2 * level, "");
        print_tree(p->args[0], level);
        printf("%*s", 2 * level, "");
        printf(";\n");
        print_tree(p->args[1], level);
    }
    else if (p->type == TRANSACTION) {
        printf("%*s", 2 * level, "");
        printf("TRANSACTION\n");
        print_tree(p->args[0], level + 1);
        print_tree(p->args[1], level + 1);
    }
    else if (p->type == READ) {
        printf("%*s", 2 * level, "");
        printf("Read\n");
        print_tree(p->args[0], level + 1);
    }
    else if (p->type == WRITE) {
        printf("%*s", 2 * level, "");
        printf("Write\n");
        print_tree(p->args[0], level + 1);
    }
    else if (p->type == LOCK) {
        printf("%*s", 2 * level, "");
        printf("Lock\n");
        print_tree(p->args[0], level + 1);
    }
    else if (p->type == READLOCK) {
        printf("%*s", 2 * level, "");
        printf("Readlock\n");
        print_tree(p->args[0], level + 1);
    }
    else if (p->type == WRITELOCK) {
        printf("%*s", 2 * level, "");
        printf("Writelock\n");
        print_tree(p->args[0], level + 1);
    }
    else if (p->type == UNLOCK) {
        printf("%*s", 2 * level, "");
        printf("Unlock\n");
        print_tree(p->args[0], level + 1);
    }
    else {
        printf("%*s", 2 * level, "");
        printf("[%c]\n", p->type);
        print_tree(p->args[0], level + 1);
        print_tree(p->args[1], level + 1);
    }
} /* print_tree */

void print_scenario(struct TreeNode* tree) {
    printf("Syntax tree:\n");
    print_tree(tree, 0);
}

void print_compact(struct TreeNode* p) {
    if (p == NULL)
        ;
    else if (p->type == ID) {
        printf("%s", get_symbol_name_unsafe(p->leaf_value));
    }
    else if (p->type == NUMBER) {
        printf("%d", p->leaf_value);
    }
    else if (p->type == ';') {
        print_compact(p->args[0]);
        printf("; ");
        print_compact(p->args[1]);
    }
    else if (p->type == TRANSACTION) {
        printf("TRANSACTION ");
        print_compact(p->args[0]);
        printf("{ ");
        print_compact(p->args[1]);
        printf("}");
    }
    else if (p->type == READ) {
        printf("Read(");
        print_compact(p->args[0]);
        printf(")");
    }
    else if (p->type == WRITE) {
        printf("Write(");
        print_compact(p->args[0]);
        printf(")");
    }
    else if (p->type == LOCK) {
        printf("Lock(");
        print_compact(p->args[0]);
        printf(")");
    }
    else if (p->type == READLOCK) {
        printf("Readlock(");
        print_compact(p->args[0]);
        printf(")");
    }
    else if (p->type == WRITELOCK) {
        printf("Writelock(");
        print_compact(p->args[0]);
        printf(")");
    }
    else if (p->type == UNLOCK) {
        printf("Unlock(");
        print_compact(p->args[0]);
        printf(")");
    }
    else if (p->type == '=') {
        print_compact(p->args[0]);
        printf(" = ");
        print_compact(p->args[1]);
    }
    else {
        printf("(");
        print_compact(p->args[0]);
        printf(" %c ", p->type);
        print_compact(p->args[1]);
        printf(")");
    }
} /* print_compact */
